#! /usr/bin/wish
#
# UPDEPON.TCL : 思考ゲーム「アップでぽん」
#
#               Copyright (C) 1999-2005,2014,2019 Makoto Hiroi
#
# 盤面の表示
# 315 * 315, 5 -- 63 * 63, 7 -- 45 * 45, 9 -- 35 * 35
#
# 大域変数
# balls()   : 玉の図形 id を格納
# place()   : 玉を置ける場所を示す図形 id を格納
# back      : 背景（図形を隠すために使用する）
# action    : 0 : 先手, 1 : 後手
# color     : 駒の色
# board_size: 盤の大きさ
# message   : メッセージ表示用ラベル
# score     : 玉数表示用ラベル
# play_flag : ゲーム進行用フラグ 1 :先手（赤）, 2 : 後手（青）
# red_ball  : 赤玉の残り（DLL 側でセット）
# blue_ball : 青玉の残り（DLL 側でセット）
#
# updepon.so 内のコマンド
#   think $level $turn : 思考
#   search_place       : ボールを置ける場所をリストで返す
#   put_ball $n $c     : ボールを置く
#                        その後に積みあがるボールの種類と場所をリストで返す
#   init_game $size    : ゲームの初期化
#

# 思考ルーチンロード
load ./updepon.so

# ヘルプファイルの表示
proc help {} {
    global argv0
    if {![winfo exist .t0]} {
        toplevel .t0
        wm title .t0 "Updepon Help"
        text .t0.text -yscrollcommand ".t0.scroll set"
        scrollbar .t0.scroll -command ".t0.text yview"
	pack .t0.scroll -side right -fill y
	pack .t0.text -side left
	# 起動パスを取得
	set path [file dirname $argv0]
	# ファイルの読み込み
	set f [open "$path/updepon.txt" r]
	while {![eof $f]} {
	    .t0.text insert end [read $f 1000]
	}
	close $f
	.t0.text configure -state disabled
    }
}

# 座表計算
proc get_zahyou_sub {d l c w s} {
    if {$d & 0x01} {
	set xs [expr $w * (($s - $d)/2) + 10]
	set ys [expr $w * (($s - $d)/2) + 10]
    } else {
	set xs [expr $w * (($s - $d)/2) + ($w / 2) + 10]
	set ys [expr $w * (($s - $d)/2) + ($w / 2) + 10]
    }
    incr xs [expr $c * $w]
    incr ys [expr $l * $w]
    set  xe [expr $xs + $w - 1]
    set  ye [expr $ys + $w - 1]
    return [list $xs $ys $xe $ye]
}


# 番号から段数、行、列を求め、表示する位置を決定
proc get_zahyou {n} {
    global board_size
    for {set d 1} {$d <= 9} {incr d} {
	set b [expr $d * $d]
	if {$b > $n} {
	    set l [expr $n / $d]
	    set c [expr $n % $d]
	    switch $board_size {
		5 {
		    set r [get_zahyou_sub $d $l $c 63 5]
		}
		7 {
		    set r [get_zahyou_sub $d $l $c 45 7]
		}
		9 {
		    set r [get_zahyou_sub $d $l $c 35 9]
		}
	    }
	    return $r
	}
	set n [expr $n - $b]
    }
}

# 図形の初期化
proc init_graph {} {
    global balls place back
    set back [.c0 create rectangle 0 0 335 335 -outline green4 -fill green4]
    for {set i 0} {$i < 285} {incr i} {
	set balls($i) [.c0 create oval 0 0 63 63]
        .c0 lower $balls($i)
	set place($i) -1
    }
}

# 玉を描く
proc draw_ball {n c} {
    global balls place
    set zl [get_zahyou $n]
    set x1 [lindex $zl 0]
    set y1 [lindex $zl 1]
    set x2 [lindex $zl 2]
    set y2 [lindex $zl 3]
    # 座標の変更
    .c0 coords $balls($n) $x1 $y1 $x2 $y2 
    # 色の変更
    .c0 itemconfigure $balls($n) -outline "dark$c" -fill $c -width 2.0
    .c0 raise $balls($n)
    # place が表示されていたら消去する
    if {$place($n) != -1} {
	.c0 delete $place($n)
	set place($n) -1
    }
    update
}


# ボールを置ける場所を表示
proc draw_place {} {
    global board_size place
    switch $board_size {
	5 { set w [expr 63 / 4]	}
	7 { set w [expr 45 / 4] }
	9 { set w [expr 35 / 4] }
    }
    foreach n [search_place] {
	if {$place($n) == -1} {
	    # まだ表示されていない
	    set zl [get_zahyou $n]
	    set x1 [expr [lindex $zl 0] + $w]
	    set y1 [expr [lindex $zl 1] + $w]
	    set x2 [expr [lindex $zl 2] - $w]
	    set y2 [expr [lindex $zl 3] - $w]
	    set place($n) [.c0 create oval $x1 $y1 $x2 $y2 \
		               -outline black -fill black]
	    # バインドの設定
	    .c0 bind $place($n) <Button-1> "put_ball_human $n"
	}
    }
}

# メッセージの表示
proc display_message {m} {
    global message mes_table
    set message $mes_table($m)
    update
    after 500
}

# スコアの表示
proc display_score {} {
    global score red_ball blue_ball
    set score [format "赤玉 %3d 個 : 青玉 %3d 個" $red_ball $blue_ball]
    update
}

# place をフラッシュさせる
proc flash_place {n} {
    global place
    if {$place($n) != -1} {
	foreach i {0 1} {
	    .c0 itemconfigure $place($n) -fill white
	    update
	    after 200
	    .c0 itemconfigure $place($n) -fill yellow
	    update
	    after 200
	}
    }
}

# 玉を置く
proc put_ball_sub {ball_list} {
    global color
    set m [llength $ball_list]
    set i 0
    flash_place [lindex $ball_list $i]
    while {$i < $m} {
	set j [lindex $ball_list $i]
	incr i
	set c [lindex $ball_list $i]
        incr i
	draw_ball $j $color($c)
	update
	after 100
    }
    display_score
}

# place を全て削除する
proc delete_all_place {} {
    global place
    for {set i 0} {$i < 285} {incr i} {
	if {$place($i) != -1} {
	    .c0 delete $place($i)
	    set place($i) -1
	}
    }
}

# 終了ダイアログ    
proc game_over {} {
    global play_flag red_ball blue_ball action back
    if {$red_ball == $blue_ball} {
	set m "引き分けです"
    } elseif {($action == 0 && $blue_ball > 0) || ($action == 1 && $red_ball > 0)} {
	set m "あなたの勝ちです"
    } else {
	set m "私の勝ちです"
    }
    tk_messageBox -type ok -message $m
    set play_flag 0
    delete_all_place
    display_message start
}

# コンピュータ側の手番
proc put_ball_com {} {
    global level play_flag red_ball blue_ball
    display_message com_turn
    set result [think $level $play_flag]
    put_ball_sub $result 
    # 終了チェック
    if {$red_ball == 0 || $blue_ball == 0} {
	game_over
	return
    }
    # 手番を移す
    if {$play_flag == 1} {
	set play_flag 2
    } else {
	set play_flag 1
    }
    draw_place
    display_message my_turn
}

# 人間側の指し手
proc put_ball_human {n} {
    global play_flag action red_ball blue_ball
    # 自分の番か？
    if {[expr $play_flag - 1] != $action} return 
    # 玉の位置と種類がリストで返される
    set bl [put_ball $n $play_flag]
    # 手番を移す（２重入力の防止）
    if {$play_flag == 1} {
	set play_flag 2
    } else {
	set play_flag 1
    }
    # 玉を描画
    put_ball_sub $bl 
    # 終了チェック
    if {$red_ball == 0 || $blue_ball == 0} {
	game_over
	return
    }
    draw_place
    display_message com_turn
    put_ball_com
}


# 中断ダイアログ
proc break_dialog {} {
    set ans [tk_messageBox -type yesno -icon question \
	                   -message "今のゲームを中断しますか？"]
    return [expr {($ans == "no") ? 0 : 1}]
}

# ゲームの開始
proc new_game {} {
    global play_flag action board_size size back red_ball blue_ball
    if {$play_flag != 0} {
	if ![break_dialog] return
	set play_flag 0
	delete_all_place
    }
    if {$play_flag == 0 } {
	# ゲーム開始
	.c0 raise $back
	set board_size $size
	init_game $board_size
	draw_place
	# 中立の玉を描く
	draw_ball 0 green
	display_score
	set play_flag 1
	if $action {
	    put_ball_com
	} else {
	    display_message my_turn
	}
    }
}

# サイズの変更
proc change_size {} {
    global play_flag size board_size
    if {$play_flag == 0} {
      set board_size $size
    }
}

# グローバル変数の初期化
proc init_global {} {
    global level action mes_table play_flag message board_size color size
    set level 1
    set board_size 5
    set size 5
    set action 0
    set play_flag 0
    set color(1) red
    set color(2) blue
    set color(3) purple
    set mes_table(start)    "メニュー Games の Start でゲーム開始！"
    set mes_table(my_turn)  "あなたの手番です"
    set mes_table(com_turn) "わたしの手番です"
    set message $mes_table(start)
}

# ********** メニューの設定 **********
menu .m -type menubar
. configure -menu .m -bg darkgreen
.m add cascade -label "Games" -under 0 -menu .m.m1
.m add cascade -label "Level" -under 0 -menu .m.m2
.m add cascade -label "Size"  -under 0 -menu .m.m3
#.m add command -label "Hint"  -under 0 -command "hint"
.m add command -label "Help"  -under 0 -command "help"

menu .m.m1 -tearoff no
.m.m1 add command -label "Start" -under 0 -command "new_game"
.m.m1 add separator
.m.m1 add radiobutton -label "先手" -variable action -value 0
.m.m1 add radiobutton -label "後手" -variable action -value 1
.m.m1 add separator
.m.m1 add command -label "Exit" -under 0 -command "exit"
menu .m.m2 -tearoff no
for {set i 1} {$i < 7} {incr i} {
  .m.m2 add radiobutton -label "Level $i" -variable level -value $i
}
menu .m.m3 -tearoff no
.m.m3 add radiobutton -label "5 * 5" -variable size -value 5 -command "change_size"
.m.m3 add radiobutton -label "7 * 7" -variable size -value 7 -command "change_size"
.m.m3 add radiobutton -label "9 * 9" -variable size -value 9 -command "change_size"

# 画面の表示
option add *font "{Noto Sans Mono CJK JP} 12"
canvas .c0 -width 335 -height 335 -bg darkgreen

# メッセージ表示用ラベル
label .lm -textvariable message -bg darkgreen -fg yellow
# スコア表示用
label .ls -textvariable score -bg darkgreen -fg white

pack .ls .lm .c0 -fill x

wm title . "Updepon"
wm resizable . 0 0

init_global
init_graph
focus -force .

# end of file
